/* REXX program (what else should this be) to check if a mail */
/* passed a host listed in a DNS-based blacklist.             */
/* This program is intended for a PMMail filter with the      */
/* PROGRAM.<\yourpath\CK_Blacklist.cmd> = <returnvalue>       */
/*  whereby <returnvalue> can be                              */
/*         "confirmed"      for confirmed spam sources        */
/*         "spam source"    for above + dialup spam sources   */
/*         "Spam-Warning" for above + open relays             */
/*                                                            */
/*  You need access to a DNS server while running this script */
/*                                                            */
/* uses RexxSock library                                      */
/*                                                            */
/* Copyright 2001, Winfried Tilanus, no_spam@tilanus.com      */
/* and 2002, Lko Willms, lueko.willms@t-online.de            */
/* http://www.willms-edv.de/freesoft/                         */
/*                                                            */
/* As long as it doesn't make you rich, you may do as you     */
/* please with this program. If you do something with it that */
/* makes you rich, please send the authors some of the money. */
/*                                                            */
/* If you make some significant code enhancements, or if you  */
/* would like to become maintainer of this script, please     */
/* e-mail me at no_spam@tilanus.com                           */
/* or lueko.willms@t-online.de                                */
/*                                                            */
/*                                                            */
/* This program comes without any warrenty, if you blow up    */
/* the internet or something else with it: you are responsibe */
/* for whatever you do with it, not me, so think for your     */
/* self. If you are not able to think for your self, don't    */
/* use this program :-))                                      */
/*                                                            */
/* The program takes 2 arguments: the filename of the file to */
/* to be checked, and the name of a file where the results    */
/* are returned, which in turn are checked in the PMMail      */
/* PROGRAM filter condition.                                  */
/*   When you enter a third parameter, "debug", the script    */
/* will SAY some debug information and wait for a ENTER at END*/
/*                                                            */
/* It writes the following lines if one of the servers        */
/* that mail passed along was blacklisted, into that file:    */
/*                                                            */
/* Spam-Warning: [followed by some details]                   */
/*                                                            */
/* The list of blacklists to use can be changed below. The    */
/* first line (blacklists.0) indicates the number of          */
/* blacklists to use. Add for each blacklist a line after it. */
/* (If you don't understand how, THEN ask somebody else, but  */
/* I prefer it if you don't ask me)                           */

/* The mechanism used relies on those blacklisting domains to */
/* register the IP addresses being blacklisted in reversed    */
/* order as subdomains of their own domain, similar as        */
/* IP-addresses are being registered in DNS for reverse       */
/* lookup as e.g. "192.10.22.64.in-addr.arpa", when the       */
/* actual IP-address is "64.22.10.192".                       */
/* In a similar vein, the blacklisters listed below would     */
/* register the above IP-address in their blacklist like      */
/* this: "192.10.22.64.relays.ordb.org", for example.         */
/* Their DNS-server THEN returns for a normal DNS address     */
/* query (Get Host address by its name) a bogus IP-address    */
/* which is just a status indicator for that blacklist.       */
/*                                                            */
/*------------------------------------------------------------*/

  blacklist.0 = 4    /* Number of blacklisters; change this if you add or delete a blacklister */
  blacklist.1 = "inputs.orbz.org"
  blacklist.2 = "relays.ordb.org"
  blacklist.3 = "relays.osirusoft.com"
  blacklist.4 = "bl.spamcop.net"

/* It should not be nessecery to change anything below this,  */
/* but if you really think it is neat or cool to do so:       */
/* GO AHEAD!                                                  */

/* load the libraries                                         */

  call rxfuncadd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
  call SysLoadFuncs

  call RxFuncAdd "SockLoadFuncs","RxSock","SockLoadFuncs"
  call SockLoadFuncs

/* get the filenames from the command line parameters ('bestand' is Dutch for file, 'ergebnis' is German for "result")     */
  ARG bestand ergebnis debug .

  IF debug = "DEBUG" THEN
    SAY bestand ergebnis

  chunk.0 = 4
  DO i = 1 TO chunk.0
    chunk.i = ""
  END /* do */


/* ----------------------------------- first step: Reading in Header lines and extracting some information from there --- */
  rc = STREAM(bestand, "C", "OPEN READ")
  fromLine = ""
  messageID = ""
  subjectLine = ""
  istReceiveHeader = 0
  FromDomains.0 = 0
  AnzahlReceiveZeilen = 0
  ReceivedHeader = ""
  headerZeile = LINEIN(bestand)
  DO until length(headerZeile)=0
   /* debug:    SAY headerZeile  */
   headerLabel = TRANSLATE(WORD(headerZeile, 1))
   SELECT
     WHEN headerLabel = "RECEIVED:"  THEN
       DO
        istReceiveHeader = 1
        ReceivedHeader = ""
       END
     WHEN headerLabel = "FROM:" THEN
       fromLine = headerZeile
     WHEN headerLabel = "MESSAGE-ID:" THEN
       messageID = headerZeile
     WHEN headerLabel = "SUBJECT:" THEN
       subjectLine = headerZeile
     OTHERWISE
       NOP
   END /* SELECT  */
   IF istReceiveHeader = 1 THEN
     DO
     IF debug = "DEBUG" THEN
        SAY "Checking " headerZeile
     ReceivedHeader = ReceivedHeader || headerZeile || " "
   END /* IF istReceiveHeader = 1 */
   headerZeile = LINEIN(bestand)
   IF LEFT(headerZeile,1) <> " " THEN
      DO
      istReceiveHeader = 0
      IF LENGTH(ReceivedHeader) > 0 THEN
        DO
         klammerBYpos = POS(")BY ", ReceiveHeader)
         IF klammerBYpos > 0 THEN
           ReceiveHeader = INSERT(" ", ReceiveHeader, klammerBYpos)
         PARSE UPPER VAR ReceivedHeader . " FROM " FromDomainAngabe " BY " .
         IF LENGTH(FromDomainAngabe) > 0 THEN
           DO
           AnzahlReceiveZeilen = AnzahlReceiveZeilen + 1
           FromDomains.AnzahlReceiveZeilen = STRIP(FromDomainAngabe,"B")
           FromDomains.0 = AnzahlReceiveZeilen
           IF debug = "DEBUG" THEN
             SAY "FromDomain Nr. " AnzahlReceiveZeilen FromDomains.AnzahlReceiveZeilen
           END
         ReceivedHeader = ""
        END
      END /* Received: Line is at END  */
  END /* DO  Headerlines einlesen */
  rc = STREAM(bestand, "C", "CLOSE")

/* ----------------------------------- second step: Recording some header information in the results file  --- */
  rc = STREAM(ergebnis,"C","OPEN WRITE")
  DO
   labelWidth = 14
   zeitAnfang = LEFT(DATE("S"),4)||"-"||SUBSTR(DATE("S"),5,2)||"-"||RIGHT(DATE("S"), 2)||" "||TIME()
   accountINIstring = STRIP(CHARIN(LEFT(bestand, POS("\", bestand))||"ACCT.INI", X2D("0201"), 256),"L",D2C(0))
   accountName = LEFT(accountINIstring, POS(D2C(0), accountINIstring) - 1)
   rc = LINEOUT(ergebnis, zeitAnfang)
   rc = LINEOUT(ergebnis, "Account: "||accountName)
   rc = LINEOUT(ergebnis, bestand)
   IF LENGTH(fromLine) > 0 THEN
      rc = LINEOUT(ergebnis, fromLine)
   IF LENGTH(subjectLine) > 0 THEN
      rc = LINEOUT(ergebnis, subjectLine)
   IF LENGTH(messageID) > 0 THEN
      rc = LINEOUT(ergebnis, messageID)
  END
  /* Note: File "ergebnis" is still open; it is being closed only in the fourth step, since the results of the next step may be written to it */

/* ----------------------------------- third step: Checking the IP-Addresses from the Received: Headers against the selected blacklists --- */
   entries = 0         /* in the stem of blacklisted IP-Adresses */
   DO i = FromDomains.0 TO 1 BY -1
     IF debug = "DEBUG" THEN
        SAY "Jetzt Adressen extrahieren aus: " FromDomains.i
     SELECT
        WHEN POS("[",FromDomains.i) > 0 THEN
          DO
          startParseAt = LASTPOS("[",FromDomains.i)
          PARSE VAR FromDomains.i =(startParseAt) "[" chunk.1 "." chunk.2 "." chunk.3 "." chunk.4 "]" .
          END
        WHEN POS("(",FromDomains.i) > 0 THEN
          DO
          startParseAt = LASTPOS("(",FromDomains.i)
          PARSE VAR FromDomains.i =(startParseAt) "(" chunk.1 "." chunk.2 "." chunk.3 "." chunk.4 ")" .
          END
        OTHERWISE
          DO
          chunk.1 = ""
          chunk.2 = ""
          chunk.3 = ""
          chunk.4 = ""
          END
     END /* Select */
     IF (DATATYPE(chunk.1) = "NUM" & DATATYPE(chunk.2) = "NUM" & DATATYPE(chunk.3) = "NUM" & DATATYPE(chunk.4) = "NUM") & chunk.1||"."||chunk.2||"."||chunk.3||"."||chunk.4 <> "127.0.0.1" THEN
       DO
        DO listno=1 to blacklist.0
          IF debug = "DEBUG" THEN
             SAY "Querying " chunk.4 || "." || chunk.3 || "." || chunk.2 || "." || chunk.1 || "." || blacklist.listno
          rc=SockGetHostByName(chunk.4 || "." || chunk.3 || "." || chunk.2 || "." || chunk.1 || "." || blacklist.listno||".", 'host.')
          IF rc=1 & (host.addr="127.0.0.2" | host.addr="127.0.0.3" | host.addr="127.0.0.4") THEN
            DO
             /* You may change the above condition, e.g., to detect only "Confirmed Spam Source", and not "Open Relay"  */
             entries = entries + 1
             IF host.addr = "127.0.0.2" THEN status = "Open Relay"
             IF host.addr = "127.0.0.3" THEN status = "Dialup Spam Source"
             IF host.addr = "127.0.0.4" THEN status = "Confirmed Spam Source"
             entry.entries = "Spam-Warning: "||status||" by "||blacklist.listno||" for IP-A. in "||D2C(34)||FromDomains.i||D2C(34)
             IF debug = "DEBUG" THEN
               SAY entry.entries
            END /* IF bad host address returned */
        END /* DO schleife ber blacklists */
      END /* IF IP-Adresse gefunden  */
   END /* DO Schleife ber FromDomains.0  */


/* ----------------------------------- fourth step: if some of the passing hosts are blacklisted, record this in results file and log file --- */
  entry.0 = entries
  zeitEnde = LEFT(DATE("S"),4)||"-"||SUBSTR(DATE("S"),5,2)||"-"||RIGHT(DATE("S"), 2)||" "||TIME()
  DO i = 1 TO entries
    rc = LINEOUT(ergebnis, entry.i)
  END /* do */
  rc = LINEOUT(ergebnis, zeitEnde)
  rc = STREAM(ergebnis,"C","CLOSE")
  /* result file is done, now record the information in the log file  */

  IF entries > 0 THEN
    DO
    PARSE SOURCE . . programmName
    programmFileName = FILESPEC("N", programmName)
    logFileName = LEFT(programmFileName, LASTPOS(".", programmFileName))||"LOG"
    spamReport = FILESPEC("D",programmName)||FILESPEC("P",programmName)||logFileName          /* typically: "CK_blacklist.LOG" */
    erg = STREAM(spamReport, "C", "OPEN WRITE")
    DO WHILE erg <> "READY:"
      CALL SysSleep 1
      erg = STREAM(spamReport, "C", "OPEN WRITE")
    END /* do */
    IF debug = "DEBUG" THEN
      SAY spamReport erg
    rc = LINEOUT(spamReport, zeitAnfang)
    rc = LINEOUT(spamReport, "Account: "||accountName)
    rc = LINEOUT(spamReport, bestand)
    IF LENGTH(fromLine) > 0 THEN
      rc = LINEOUT(spamReport, fromLine)
    IF LENGTH(subjectLine) > 0 THEN
      rc = LINEOUT(spamReport, subjectLine)
    IF LENGTH(messageID) > 0 THEN
      rc = LINEOUT(spamReport, messageID)
    DO i = 1 TO entries
      rc = LINEOUT(spamReport, entry.i)
    END /* do */
    rc = LINEOUT(spamReport, zeitEnde)
    rc = LINEOUT(spamReport, "")
    rc = STREAM(spamReport,"C","CLOSE")
  END

/*  for debugging: stop the execution of the program until the user hits ENTER */
  IF debug = "DEBUG" THEN
    DO
     rc = CHAROUT(,"Weiter mit ENTER")
     PULL dummy
    END

EXIT

/* END of program */


